package laplace;

import laplace.graphics.GraphicsPanel;
import laplace.model.BoundaryCondition.Direction;
import laplace.model.BoundaryCondition;
import laplace.model.CommunicationBoundary;
import laplace.model.AreaFunction;
import java.util.ArrayList;
import java.util.Collection;
import javax.swing.JComponent;
import javax.swing.JFrame;
import muscle.core.kernel.CAController;
import muscle.core.kernel.Submodel;
import muscle.core.model.Timestamp;

/**
 * @author Joris Borgdorff
 */
public class TemperatureKernel extends Submodel {
	private CommunicationBoundary remoteWest, remoteEast, remoteNorth, remoteSouth;
	protected int nx, ny;
	protected int dx; // dy=dx
	protected int originX, originY;
	private String name;
	private boolean hasRemoteEast, hasRemoteWest;
	private boolean hasRemoteNorth, hasRemoteSouth;

	private long nano;
	private int iteration;
	
	protected double[][] data, nextData;
	protected BoundaryCondition north, south, west, east;
	protected BoundaryCondition initialCondition;
	protected AreaFunction areaFunction;
	protected GraphicsPanel panel;
	
	/**
	 * Initializes the boundary conditions and the area calculation.
	 *
	 * It does so with an averaging area function, a sine function for the top, cosine for the bottom
	 * and a zero left and right function. If "remoteEast/West" property is true, the left or right will
	 * be communicated.
	 */
	protected void initFunctions() {
		// Set to 0
		initialCondition = new BoundaryCondition(Direction.Any, nx, ny);

		// Does simple averaging of the four adjacent cells
		areaFunction = new AreaFunction(nx, ny);

		if (hasRemoteNorth) {
			north = remoteNorth = new CommunicationBoundary(Direction.North, nx, ny, data);
		} else {
			north = new BoundaryCondition(Direction.North, nx, ny) {
				public double get(int x, int y, int step) {
					return Math.sin((x + originX) * Math.PI / nx);
				}
			};
		}
		if (hasRemoteSouth) {
			south = remoteSouth = new CommunicationBoundary(Direction.South, nx, ny, data);
		} else {
			south = new BoundaryCondition(Direction.South, nx, ny) {
				public double get(int x, int y, int step) {
					return Math.cos((x + originX) * Math.PI / nx + Math.PI);
				}
			};
		}
		if (hasRemoteWest) {
			west = remoteWest = new CommunicationBoundary(Direction.West, nx, ny, data);
		} else {
			west = new BoundaryCondition(Direction.West, nx, ny);
		}
		if (hasRemoteEast) {
			east = remoteEast = new CommunicationBoundary(Direction.East, nx, ny, data);
		} else {
			east = new BoundaryCondition(Direction.East, nx, ny);
		}
	}
	
	/** Initialize the boundary conditions and show the displays */
	@Override
	protected Timestamp init(Timestamp prevTime) {
		// Get the properties from the CxA file
		nx = getIntProperty("nx");
		ny = getIntProperty("ny");
		originX = hasProperty("originx") ? getIntProperty("originx") : 0;
		originY = hasProperty("originy") ? getIntProperty("originy") : 0;
		dx = hasProperty("dx") ? getIntProperty("dx") : 1;
		name = getProperty("name");
		hasRemoteEast = hasInstanceProperty("remoteEast") && getBooleanProperty("remoteEast");
		hasRemoteWest = hasInstanceProperty("remoteWest") && getBooleanProperty("remoteWest");
		hasRemoteNorth = hasInstanceProperty("remoteNorth") && getBooleanProperty("remoteNorth");
		hasRemoteSouth = hasInstanceProperty("remoteSouth") && getBooleanProperty("remoteSouth");

		// Initialize internal data structures
		data = new double[nx][ny];
		nextData = new double[nx][ny];

		initFunctions();

		// Start a window
		panel = new GraphicsPanel(nx, ny, dx);
		panel.show(name, originX*dx, originY*dx);

		iteration = -1;
		
		// fill with initial conditions
		for (int x = 0; x < nx; x++) {
			for (int y = 0; y < ny; y++) {
				BoundaryCondition bc;
				
				if      (north.applies(x, y)) bc = north;
				else if (south.applies(x, y)) bc = south;
				else if (east.applies(x, y))  bc = east;
				else if (west.applies(x, y))  bc = west;
				else                          bc = initialCondition;

				data[x][y] = bc.get(x, y, iteration);
			}
		}
		panel.paintAndWait(data);
		
		nano = System.nanoTime();
		
		return super.init(prevTime);
	}
	
	@Override
	protected void intermediateObservation() {
		if (hasRemoteNorth)
			this.<double[]>out("northBoundary").send(remoteNorth.getLocalBoundary(data));
		if (hasRemoteSouth)
			this.<double[]>out("southBoundary").send(remoteSouth.getLocalBoundary(data));
		if (hasRemoteWest)
			this.<double[]>out("westBoundary").send(remoteWest.getLocalBoundary(data));
		if (hasRemoteEast)
			this.<double[]>out("eastBoundary").send(remoteEast.getLocalBoundary(data));
	}

	// iterate	
	protected void solvingStep() {
		iteration++;
		areaFunction.updateData(data);
		if (hasRemoteNorth)
			remoteNorth.updateData(data, this.<double[]>in("remoteNorth").receive());
		if (hasRemoteSouth)
			remoteSouth.updateData(data, this.<double[]>in("remoteSouth").receive());
		if (hasRemoteWest)
			remoteWest.updateData(data, this.<double[]>in("remoteWest").receive());
		if (hasRemoteEast)
			remoteEast.updateData(data, this.<double[]>in("remoteEast").receive());
		
		for (int x = 0; x < nx; x++) {
			for (int y = 0; y < ny; y++) {
				BoundaryCondition bc;
				if      (north.applies(x, y)) bc = north;
				else if (south.applies(x, y)) bc = south;
				else if (east.applies(x, y))  bc = east;
				else if (west.applies(x, y))  bc = west;
				else                          bc = areaFunction;
		
				nextData[x][y] = bc.get(x, y, iteration);
			}
		}

		swapDataContainers();

		panel.paintAndWait(data);

		if (iteration % 100 == 0) {
			log("iteration " + iteration + " of " + getIntProperty("max_timesteps") + " done at runtime " + (System.nanoTime() - nano) / 1000000000d);		
		}
	}

	private void swapDataContainers() {
		// swap data containers
		double[][] tmpData = data;
		data = nextData;
		nextData = tmpData;
	}
}
